# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation

import numpy as np
import matplotlib.pyplot as plt


try:
    from mpi4py.futures import MPIPoolExecutor
except ImportError:
    pass


class Optimization:
    """
    Optimization is the base optimization class for
    `~.tools.optimization.scipy` subclasses. Contains some common
    methods and properties that can be used by the individual optimization
    classes.
    """

    def __init__(self, fi):
        """
        Initializes an Optimization object by assigning a
        FlorisInterface object.

        Args:
            fi (:py:class:`~.tools.floris_interface.FlorisInterface`):
                Interface used to interact with the Floris object.
        """
        self.fi = fi

    # Private methods

    def _reinitialize(self):
        pass

    def _norm(self, val, x1, x2):
        return (val - x1) / (x2 - x1)

    def _unnorm(self, val, x1, x2):
        return np.array(val) * (x2 - x1) + x1

    # Properties

    @property
    def nturbs(self):
        """
        Number of turbines in the :py:class:`~.farm.Farm` object.

        Returns:
            int
        """
        self._nturbs = len(self.fi.floris.farm.turbine_map.turbines)
        return self._nturbs



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation

import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import minimize

from .optimization import Optimization


class LayoutOptimization(Optimization):
    """
    Layout is a subclass of the
    :py:class:`~.tools.optimization.scipy.optimization.Optimization` class
    that is used to perform layout optimization.
    """

    def __init__(
        self,
        fi,
        boundaries,
        wd,
        ws,
        freq,
        AEP_initial,
        x0=None,
        bnds=None,
        min_dist=None,
        opt_method="SLSQP",
        opt_options=None,
    ):
        """
        Instantiate LayoutOptimization object with a FlorisInterface object and
        assign parameter values.

        Args:
            fi (:py:class:`~.tools.floris_interface.FlorisInterface`):
                Interface used to interact with the Floris object.
            boundaries (iterable(float, float)): Pairs of x- and y-coordinates
                that represent the boundary's vertices (m).
            wd (np.array): An array of wind directions (deg).
            ws (np.array): An array of wind speeds (m/s).
            freq (np.array): An array of the frequencies of occurance
                correponding to each pair of wind direction and wind speed
                values.
            AEP_initial (float): The initial Annual Energy
                Production used for normalization in the optimization (Wh)
                (TODO: Is Watt-hours the correct unit?).
            x0 (iterable, optional): The initial turbine locations,
                ordered by x-coordinate and then y-coordiante
                (ie. [x1, x2, ..., xn, y1, y2, ..., yn]) (m). If none are
                provided, x0 initializes to the current turbine locations.
                Defaults to None.
            bnds (iterable, optional): Bounds for the optimization
                variables (pairs of min/max values for each variable (m)). If
                none are specified, they are set to the min. and max. of the
                boundaries iterable. Defaults to None.
            min_dist (float, optional): The minimum distance to be
                maintained between turbines during the optimization (m). If not
                specified, initializes to 2 rotor diameters. Defaults to None.
            opt_method (str, optional): The optimization method used by
                scipy.optimize.minize. Defaults to 'SLSQP'.
            opt_options (dict, optional): Optimization options used by
                scipy.optimize.minize. If none are specified, they are set to
                {'maxiter': 100, 'disp': True, 'iprint': 2, 'ftol': 1e-9}.
                Defaults to None.
        """
        super().__init__(fi)
        self.epsilon = np.finfo(float).eps

        if opt_options is None:
            self.opt_options = {"maxiter": 100, "disp": True, "iprint": 2, "ftol": 1e-9}

        self.reinitialize_opt(
            boundaries=boundaries,
            wd=wd,
            ws=ws,
            freq=freq,
            AEP_initial=AEP_initial,
            x0=x0,
            bnds=bnds,
            min_dist=min_dist,
            opt_method=opt_method,
            opt_options=opt_options,
        )

    # Private methods

    def _AEP_layout_opt(self, locs):
        locs_unnorm = [
            self._unnorm(valx, self.bndx_min, self.bndx_max)
            for valx in locs[0 : self.nturbs]
        ] + [
            self._unnorm(valy, self.bndy_min, self.bndy_max)
            for valy in locs[self.nturbs : 2 * self.nturbs]
        ]
        self._change_coordinates(locs_unnorm)
        AEP_sum = self._AEP_loop_wd()
        return -1 * AEP_sum / self.AEP_initial

    def _AEP_single_wd(self, wd, ws, freq):
        self.fi.reinitialize_flow_field(wind_direction=[wd], wind_speed=[ws])
        self.fi.calculate_wake()

        turb_powers = [turbine.power for turbine in self.fi.floris.farm.turbines]
        return np.sum(turb_powers) * freq * 8760

    def _AEP_loop_wd(self):
        AEP_sum = 0

        for i in range(len(self.wd)):
            self.fi.reinitialize_flow_field(
                wind_direction=[self.wd[i]], wind_speed=[self.ws[i]]
            )
            self.fi.calculate_wake()

            AEP_sum = AEP_sum + self.fi.get_farm_power() * self.freq[i] * 8760
        return AEP_sum

    def _change_coordinates(self, locs):
        # Parse the layout coordinates
        layout_x = locs[0 : self.nturbs]
        layout_y = locs[self.nturbs : 2 * self.nturbs]
        layout_array = [layout_x, layout_y]

        # Update the turbine map in floris
        self.fi.reinitialize_flow_field(layout_array=layout_array)

    def _space_constraint(self, x_in, min_dist):
        x = np.nan_to_num(x_in[0 : self.nturbs])
        y = np.nan_to_num(x_in[self.nturbs :])

        dist = [
            np.sqrt((x[i] - x[j]) ** 2 + (y[i] - y[j]) ** 2)
            for i in range(self.nturbs)
            for j in range(self.nturbs)
            if i != j
        ]

        # dist = []
        # for i in range(self.nturbs):
        #     for j in range(self.nturbs):
        #         if i != j:
        #             dist.append(np.sqrt( (x[i]-x[j])**2 + (y[i]-y[j])**2))

        return np.min(dist) - self._norm(min_dist, self.bndx_min, self.bndx_max)

    def _distance_from_boundaries(self, x_in, boundaries):
        # x = self._unnorm(x_in[0:self.nturbs], self.bndx_min, self.bndx_max)
        # y = self._unnorm(x_in[self.nturbs:2*self.nturbs], \
        #                  self.bndy_min, self.bndy_max)
        x = x_in[0 : self.nturbs]
        y = x_in[self.nturbs : 2 * self.nturbs]

        dist_out = []

        for k in range(self.nturbs):
            dist = []
            in_poly = self._point_inside_polygon(x[k], y[k], boundaries)

            for i in range(len(boundaries)):
                boundaries = np.array(boundaries)
                p1 = boundaries[i]
                if i == len(boundaries) - 1:
                    p2 = boundaries[0]
                else:
                    p2 = boundaries[i + 1]

                px = p2[0] - p1[0]
                py = p2[1] - p1[1]
                norm = px * px + py * py

                u = (
                    (x[k] - boundaries[i][0]) * px + (y[k] - boundaries[i][1]) * py
                ) / float(norm)

                if u <= 0:
                    xx = p1[0]
                    yy = p1[1]
                elif u >= 1:
                    xx = p2[0]
                    yy = p2[1]
                else:
                    xx = p1[0] + u * px
                    yy = p1[1] + u * py

                dx = x[k] - xx
                dy = y[k] - yy
                dist.append(np.sqrt(dx * dx + dy * dy))

            dist = np.array(dist)
            if in_poly:
                dist_out.append(np.min(dist))
            else:
                dist_out.append(-np.min(dist))

        dist_out = np.array(dist_out)

        return np.min(dist_out)

    def _point_inside_polygon(self, x, y, poly):
        n = len(poly)
        inside = False

        p1x, p1y = poly[0]
        for i in range(n + 1):
            p2x, p2y = poly[i % n]
            if y > min(p1y, p2y):
                if y <= max(p1y, p2y):
                    if x <= max(p1x, p2x):
                        if p1y != p2y:
                            xinters = (y - p1y) * (p2x - p1x) / (p2y - p1y) + p1x
                        if p1x == p2x or x <= xinters:
                            inside = not inside
            p1x, p1y = p2x, p2y

        return inside

    def _generate_constraints(self):
        # grad_constraint1 = grad(self._space_constraint)
        # grad_constraint2 = grad(self._distance_from_boundaries)

        tmp1 = {
            "type": "ineq",
            "fun": lambda x, *args: self._space_constraint(x, self.min_dist),
            "args": (self.min_dist,),
        }
        tmp2 = {
            "type": "ineq",
            "fun": lambda x, *args: self._distance_from_boundaries(
                x, self.boundaries_norm
            ),
            "args": (self.boundaries_norm,),
        }

        self.cons = [tmp1, tmp2]

    def _optimize(self):
        self.residual_plant = minimize(
            self._AEP_layout_opt,
            self.x0,
            method=self.opt_method,
            bounds=self.bnds,
            constraints=self.cons,
            options=self.opt_options,
        )

        opt_results = self.residual_plant.x

        return opt_results

    def _set_opt_bounds(self):
        self.bnds = [(0.0, 1.0) for _ in range(2 * self.nturbs)]

    # Public methods

    def optimize(self):
        """
        This method finds the optimized layout of wind turbines for power
        production given the provided frequencies of occurance of wind
        conditions (wind speed, direction).

        Returns:
            opt_locs (iterable): A list of the optimized locations of each
            turbine (m).
        """
        print("=====================================================")
        print("Optimizing turbine layout...")
        print("Number of parameters to optimize = ", len(self.x0))
        print("=====================================================")

        opt_locs_norm = self._optimize()

        print("Optimization complete.")

        opt_locs = [
            [
                self._unnorm(valx, self.bndx_min, self.bndx_max)
                for valx in opt_locs_norm[0 : self.nturbs]
            ],
            [
                self._unnorm(valy, self.bndy_min, self.bndy_max)
                for valy in opt_locs_norm[self.nturbs : 2 * self.nturbs]
            ],
        ]

        return opt_locs

    def reinitialize_opt(
        self,
        boundaries=None,
        wd=None,
        ws=None,
        freq=None,
        AEP_initial=None,
        x0=None,
        bnds=None,
        min_dist=None,
        opt_method=None,
        opt_options=None,
    ):
        """
        This method reinitializes any optimization parameters that are
        specified. Otherwise, the current parameter values are kept.

        Args:
            boundaries (iterable(float, float)): Pairs of x- and y-coordinates
                that represent the boundary's vertices (m).
            wd (np.array): An array of wind directions (deg). Defaults to None.
            ws (np.array): An array of wind speeds (m/s). Defaults to None.
            freq (np.array): An array of the frequencies of occurance
                correponding to each pair of wind direction and wind speed
                values. Defaults to None.
            AEP_initial (float): The initial Annual Energy
                Production used for normalization in the optimization (Wh). If
                not specified, initializes to the AEP of the current Floris
                object. Defaults to None.
            x0 (iterable, optional): The initial turbine locations,
                ordered by x-coordinate and then y-coordiante (ie. [x1, x2, ...
                , xn, y1, y2, ..., yn] (m)). If none are provided, x0
                initializes to the current turbine locations. Defaults to None.
            bnds (iterable, optional): Bounds for the optimization
                variables (pairs of min/max values for each variable (m)). If
                none are specified, they are set to the min. and max. of the
                boundaries iterable. Defaults to None.
            min_dist (float, optional): The minimum distance to be maintained
                between turbines during the optimization (m). If not specified,
                initializes to 2 rotor diameters. Defaults to None.
            opt_method (str, optional): The optimization method for
                scipy.optimize.minize to use. If none is specified, initializes
                to 'SLSQP'. Defaults to None.
            opt_options (dict, optional): Dicitonary for setting the
                optimization options. Defaults to None.
        """
        if boundaries is not None:
            self.boundaries = boundaries
            self.bndx_min = np.min([val[0] for val in boundaries])
            self.bndy_min = np.min([val[1] for val in boundaries])
            self.bndx_max = np.max([val[0] for val in boundaries])
            self.bndy_max = np.max([val[1] for val in boundaries])
            self.boundaries_norm = [
                [
                    self._norm(val[0], self.bndx_min, self.bndx_max),
                    self._norm(val[1], self.bndy_min, self.bndy_max),
                ]
                for val in self.boundaries
            ]
        if wd is not None:
            self.wd = wd
        if ws is not None:
            self.ws = ws
        if freq is not None:
            self.freq = freq
        if AEP_initial is not None:
            self.AEP_initial = AEP_initial
        else:
            self.AEP_initial = self.fi.get_farm_AEP(self.wd, self.ws, self.freq)
        if x0 is not None:
            self.x0 = x0
        else:
            self.x0 = [
                self._norm(coord.x1, self.bndx_min, self.bndx_max)
                for coord in self.fi.floris.farm.turbine_map.coords
            ] + [
                self._norm(coord.x2, self.bndy_min, self.bndy_max)
                for coord in self.fi.floris.farm.turbine_map.coords
            ]
        if bnds is not None:
            self.bnds = bnds
        else:
            self._set_opt_bounds()
        if min_dist is not None:
            self.min_dist = min_dist
        else:
            self.min_dist = 2 * self.fi.floris.farm.turbines[0].rotor_diameter
        if opt_method is not None:
            self.opt_method = opt_method
        if opt_options is not None:
            self.opt_options = opt_options

        self._generate_constraints()

    def plot_layout_opt_results(self):
        """
        This method plots the original and new locations of the turbines in a
        wind farm after layout optimization.
        """
        locsx_old = [
            self._unnorm(valx, self.bndx_min, self.bndx_max)
            for valx in self.x0[0 : self.nturbs]
        ]
        locsy_old = [
            self._unnorm(valy, self.bndy_min, self.bndy_max)
            for valy in self.x0[self.nturbs : 2 * self.nturbs]
        ]
        locsx = [
            self._unnorm(valx, self.bndx_min, self.bndx_max)
            for valx in self.residual_plant.x[0 : self.nturbs]
        ]
        locsy = [
            self._unnorm(valy, self.bndy_min, self.bndy_max)
            for valy in self.residual_plant.x[self.nturbs : 2 * self.nturbs]
        ]

        plt.figure(figsize=(9, 6))
        fontsize = 16
        plt.plot(locsx_old, locsy_old, "ob")
        plt.plot(locsx, locsy, "or")
        # plt.title('Layout Optimization Results', fontsize=fontsize)
        plt.xlabel("x (m)", fontsize=fontsize)
        plt.ylabel("y (m)", fontsize=fontsize)
        plt.axis("equal")
        plt.grid()
        plt.tick_params(which="both", labelsize=fontsize)
        plt.legend(
            ["Old locations", "New locations"],
            loc="lower center",
            bbox_to_anchor=(0.5, 1.01),
            ncol=2,
            fontsize=fontsize,
        )

        verts = self.boundaries
        for i in range(len(verts)):
            if i == len(verts) - 1:
                plt.plot([verts[i][0], verts[0][0]], [verts[i][1], verts[0][1]], "b")
            else:
                plt.plot(
                    [verts[i][0], verts[i + 1][0]], [verts[i][1], verts[i + 1][1]], "b"
                )



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation


"""
Defines the BaseClass parent class for all models to be based upon.
"""

from abc import ABC, abstractmethod
from typing import Any, Dict, Final

import attrs
from attrs import define

from floris.type_dec import FromDictMixin
from floris.logging_manager import LoggerBase


class BaseClass(LoggerBase, FromDictMixin):
    """
    BaseClass object class. This class does the logging and MixIn class inheritance.
    """

    @classmethod
    def get_model_defaults(cls) -> Dict[str, Any]:
        """Produces a dictionary of the keyword arguments and their defaults.

        Returns
        -------
        Dict[str, Any]
            Dictionary of keyword argument: default.
        """
        return {el.name: el.default for el in attrs.fields(cls)}

    def _get_model_dict(self) -> dict:
        """Convenience method that wraps the `attrs.asdict` method. Returns the object's
        parameters as a dictionary.

        Returns
        -------
        dict
            The provided or default, if no input provided, model settings as a dictionary.
        """
        return attrs.asdict(self)


class BaseModel(BaseClass, ABC):
    """
    BaseModel is the generic class for any wake models. It defines the API required to
    create a valid model.
    """

    NUM_EPS: Final[float] = 0.001  # This is a numerical epsilon to prevent divide by zeros

    @property
    def model_string(self):
        return self.model_string

    @model_string.setter
    @abstractmethod
    def model_string(self, string):
        raise NotImplementedError("BaseModel.model_string")

    @abstractmethod
    def prepare_function() -> dict:
        raise NotImplementedError("BaseModel.prepare_function")

    @abstractmethod
    def function() -> None:
        raise NotImplementedError("BaseModel.function")



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation


import numpy as np
import matplotlib.pyplot as plt


def cluster_turbines(fi, wind_direction=None, wake_slope=0.30, plot_lines=False):
    """Separate a wind farm into separate clusters in which the turbines in
    each subcluster only affects the turbines in its cluster and has zero
    interaction with turbines from other clusters, both ways (being waked,
    generating wake), This allows the user to separate the control setpoint
    optimization in several lower-dimensional optimization problems, for
    example. This function assumes a very simplified wake function where the
    wakes are assumed to have a linearly diverging profile. In comparisons
    with the FLORIS GCH model, the wake_slope matches well with the FLORIS'
    wake profiles for a value of wake_slope = 0.5 * turbulence_intensity, where
    turbulence_intensity is an input to the FLORIS model at the default
    GCH parameterization. Note that does not include wind direction variability.
    To be conservative, the user is recommended to use the rule of thumb:
    `wake_slope = turbulence_intensity`. Hence, the default value for
    `wake_slope=0.30` should be conservative for turbulence intensities up to
    0.30 and is likely to provide valid estimates of which turbines are
    downstream until a turbulence intensity of 0.50. This simple model saves
    time compared to FLORIS.

    Args:
        fi ([floris object]): FLORIS object of the farm of interest.
        wind_direction (float): The wind direction in the FLORIS frame
        of reference for which the downstream turbines are to be determined.
        wake_slope (float, optional): linear slope of the wake (dy/dx)
        plot_lines (bool, optional): Enable plotting wakes/turbines.
        Defaults to False.

    Returns:
        clusters (iterable): A list in which each entry contains a list
        of turbine numbers that together form a cluster which
        exclusively interact with one another and have zero
        interaction with turbines outside of this cluster.
    """

    if wind_direction is None:
        wind_direction = np.mean(fi.floris.farm.wind_direction)

    # Get farm layout
    x = fi.layout_x
    y = fi.layout_y
    D = np.array([t.rotor_diameter for t in fi.floris.farm.turbines])
    n_turbs = len(x)

    # Rotate farm and determine freestream/waked turbines
    is_downstream = [False for _ in range(n_turbs)]
    x_rot = (
        np.cos((wind_direction - 270.0) * np.pi / 180.0) * x
        - np.sin((wind_direction - 270.0) * np.pi / 180.0) * y
    )
    y_rot = (
        np.sin((wind_direction - 270.0) * np.pi / 180.0) * x
        + np.cos((wind_direction - 270.0) * np.pi / 180.0) * y
    )

    if plot_lines:
        fig, ax = plt.subplots()
        for ii in range(n_turbs):
            ax.plot(
                x_rot[ii] * np.ones(2),
                [y_rot[ii] - D[ii] / 2, y_rot[ii] + D[ii] / 2],
                "k",
            )
        for ii in range(n_turbs):
            ax.text(x_rot[ii], y_rot[ii], "T%03d" % ii)
        ax.axis("equal")

    srt = np.argsort(x_rot)
    usrt = np.argsort(srt)
    x_rot_srt = x_rot[srt]
    y_rot_srt = y_rot[srt]
    affected_by_turbs = np.tile(False, (n_turbs, n_turbs))
    for ii in range(n_turbs):
        x0 = x_rot_srt[ii]
        y0 = y_rot_srt[ii]

        def wake_profile_ub_turbii(x):
            y = (y0 + D[ii]) + (x - x0) * wake_slope
            if isinstance(y, (float, np.float64, np.float32)):
                if x < (x0 + 0.01):
                    y = -np.Inf
            else:
                y[x < x0 + 0.01] = -np.Inf
            return y

        def wake_profile_lb_turbii(x):
            y = (y0 - D[ii]) - (x - x0) * wake_slope
            if isinstance(y, (float, np.float64, np.float32)):
                if x < (x0 + 0.01):
                    y = -np.Inf
            else:
                y[x < x0 + 0.01] = -np.Inf
            return y

        def determine_if_in_wake(xt, yt):
            return (yt < wake_profile_ub_turbii(xt)) & (yt > wake_profile_lb_turbii(xt))

        # Get most downstream turbine
        is_downstream[ii] = not any(
            [
                determine_if_in_wake(x_rot_srt[iii], y_rot_srt[iii])
                for iii in range(n_turbs)
            ]
        )
        # Determine which turbines are affected by this turbine ('ii')
        affecting_following_turbs = [
                determine_if_in_wake(x_rot_srt[iii], y_rot_srt[iii])
                for iii in range(n_turbs)
        ]

        # Determine by which turbines this turbine ('ii') is affected
        for aft in np.where(affecting_following_turbs)[0]:
            affected_by_turbs[aft, ii] = True

        if plot_lines:
            x1 = np.max(x_rot_srt) + 500.0
            ax.fill_between(
                [x0, x1, x1, x0],
                [
                    wake_profile_ub_turbii(x0 + 0.02),
                    wake_profile_ub_turbii(x1),
                    wake_profile_lb_turbii(x1),
                    wake_profile_lb_turbii(x0 + 0.02),
                ],
                alpha=0.1,
                color="k",
                edgecolor=None,
            )

    # Rearrange into initial frame of reference
    affected_by_turbs = affected_by_turbs[:, usrt][usrt, :]
    for ii in range(n_turbs):
        affected_by_turbs[ii, ii] = True  # Add self to turb_list_affected
    affected_by_turbs = [np.where(c)[0] for c in affected_by_turbs]

    # List of downstream turbines
    turbs_downstream = [is_downstream[i] for i in usrt]
    turbs_downstream = list(np.where(turbs_downstream)[0])

    # Initialize one cluster for each turbine and all the turbines its affected by
    clusters = affected_by_turbs

    # Iteratively merge clusters if any overlap between turbines
    ci = 0
    while ci < len(clusters):
        # Compare current row to the ones to the right of it
        cj = ci + 1
        merged_column = False
        while cj < len(clusters):
            if any([y in clusters[ci] for y in clusters[cj]]):
                # Merge
                clusters[ci] = np.hstack([clusters[ci], clusters[cj]])
                clusters[ci] = np.array(np.unique(clusters[ci]), dtype=int)
                clusters.pop(cj)
                merged_column = True
            else:
                cj = cj + 1
        if not merged_column:
            ci = ci + 1

    if plot_lines:
        ax.set_title("wind_direction = %.1f deg" % wind_direction)
        ax.set_xlim([np.min(x_rot) - 500.0, x1])
        ax.set_ylim([np.min(y_rot) - 500.0, np.max(y_rot) + 500.0])
        for ci, cl in enumerate(clusters):
            ax.plot(x_rot[cl], y_rot[cl], 'o', label='cluster %d' % ci)
        ax.legend()

    return clusters


# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

from typing import Any, Dict

from attrs import define, field
import numexpr as ne
import numpy as np

from floris.simulation import BaseModel
from floris.simulation import Farm
from floris.simulation import FlowField
from floris.simulation import Grid
from floris.simulation import Turbine


@define
class JensenVelocityDeficit(BaseModel):
    """
    The Jensen model computes the wake velocity deficit based on the classic
    Jensen/Park model :cite:`jvm-jensen1983note`.

    -   **we** (*float*): The linear wake decay constant that
        defines the cone boundary for the wake as well as the
        velocity deficit. D/2 +/- we*x is the cone boundary for the
        wake.

    References:
        .. bibliography:: /source/zrefs.bib
            :style: unsrt
            :filter: docname in docnames
            :keyprefix: jvm-
    """

    we: float = field(converter=float, default=0.05)
    model_string = "jensen"

    def prepare_function(
        self,
        grid: Grid,
        flow_field: FlowField,
        turbine: Turbine
    ) -> Dict[str, Any]:
        """
        This function prepares the inputs from the various FLORIS data structures
        for use in the Jensen model. This should only be used to 'initialize'
        the inputs. For any data that should be updated successively,
        do not use this function and instead pass that data directly to
        the model function.
        """
        kwargs = dict(
            x=grid.x,
            y=grid.y,
            z=grid.z,
            reference_rotor_diameter=turbine.rotor_diameter
        )
        return kwargs

    # @profile
    def function(
        self,
        x_i: np.ndarray,
        y_i: np.ndarray,
        z_i: np.ndarray,
        axial_induction_i: np.ndarray,
        deflection_field_i: np.ndarray,
        yaw_angle_i: np.ndarray,
        turbulence_intensity_i: np.ndarray,
        ct_i: np.ndarray,
        # enforces the use of the below as keyword arguments and adherence to the
        # unpacking of the results from prepare_function()
        *,
        x: np.ndarray,
        y: np.ndarray,
        z: np.ndarray,
        reference_rotor_diameter: float,
    ) -> None:

        # u is 4-dimensional (n wind speeds, n turbines, grid res 1, grid res 2)
        # velocities is 3-dimensional (n turbines, grid res 1, grid res 2)

        # grid.rotate_fields(flow_field.wind_directions)  # TODO: check the rotations with multiple directions or non-0/270

        # Calculate and apply wake mask
        # x = grid.x # mesh_x_rotated - x_coord_rotated

        # This is the velocity deficit seen by the i'th turbine due to wake effects from upstream turbines.
        # Indeces of velocity_deficit corresponding to unwaked turbines will have 0's
        # velocity_deficit = np.zeros(np.shape(flow_field.u_initial))

        reference_rotor_radius = reference_rotor_diameter / 2.0

        """
        dx = x - x_i
        dy = y - y_i - deflection_field_i
        dz = z - z_i

        # y = m * x + b
        boundary_line = self.we * dx + reference_rotor_radius

        # Calculate the wake velocity deficit ratios
        # Do we need to do masking here or can it be handled in the solver?
        # TODO: why do we need to slice with i:i+1 below? This became a problem when adding the wind direction dimension. Prior to that, the dimensions worked out simply with i
        c = ( reference_rotor_radius / ( reference_rotor_radius + self.we * dx + self.NUM_EPS ) ) ** 2
        # c *= ~(np.array(x - x[:, :, i:i+1] <= 0.0))  # using this causes nan's in the upstream turbine because it negates the mask rather than setting it to 0. When self.we * (x - x[:, :, i:i+1]) ) == the radius, c goes to infinity and then this line flips it to Nans rather than setting to 0.
        # c *= ~(((y - y_center) ** 2 + (z - z_center) ** 2) > (boundary_line ** 2))
        # np.nan_to_num

        # C should be 0 at the current turbine and everywhere in front of it
        downstream_mask = np.array(dx > 0.0 + self.NUM_EPS, dtype=int)
        # C should be 0 everywhere outside of the lateral and vertical bounds defined by the wake expansion parameter
        boundary_mask = np.array( np.sqrt(dy ** 2 + dz ** 2) < boundary_line, dtype=int)

        mask = np.logical_and(downstream_mask, boundary_mask)
        c[~mask] = 0.0

        velocity_deficit = 2 * axial_induction_i * c
        """

        # Numexpr - do not change below without corresponding changes above.
        dx = ne.evaluate("x - x_i")
        dy = ne.evaluate("y - y_i - deflection_field_i")
        dz = ne.evaluate("z - z_i")

        we = self.we
        NUM_EPS = JensenVelocityDeficit.NUM_EPS

        # y = m * x + b
        boundary_line = ne.evaluate("we * dx + reference_rotor_radius")

        c = ne.evaluate("( reference_rotor_radius / ( reference_rotor_radius + we * dx + NUM_EPS ) ) ** 2")

        # C should be 0 at the current turbine and everywhere in front of it
        downstream_mask = ne.evaluate("dx > 0 + NUM_EPS")

        # C should be 0 everywhere outside of the lateral and vertical bounds defined by the wake expansion parameter
        boundary_mask = ne.evaluate("sqrt(dy ** 2 + dz ** 2) < boundary_line")
        
        mask = np.logical_and(downstream_mask, boundary_mask)
        c[~mask] = 0.0
        # c = ne.evaluate("c * downstream_mask * boundary_mask")

        velocity_deficit = ne.evaluate("2 * axial_induction_i * c")

        return velocity_deficit



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation

import logging
from datetime import datetime

import coloredlogs


# Global variables for logging
LOG_TO_CONSOLE = True
CONSOLE_LEVEL = "INFO"
LOG_TO_FILE = False
FILE_LEVEL = "INFO"


def configure_console_log(enabled=True, level="INFO"):
    """
    Sets whether the log statements are displayed in the console logging, and,
    if enabled, the log level to use. If not explicitly configured, console
    logging is ON at the INFO level.

    Args:
        enabled (bool, optional): Whether to enable console logging.
            Defaults to True.
        level (str, optional): If `enabled` is True, sets the level that the
            logging module displays. This level is the minimum and all
            messages at a higher level are included. Valid values are

                - CRITICAL
                - ERROR
                - WARNING
                - INFO
                - DEBUG

            Defaults to "INFO".
    """
    global LOG_TO_CONSOLE
    global CONSOLE_LEVEL
    LOG_TO_CONSOLE = enabled
    CONSOLE_LEVEL = level
    _setup_logger()


def configure_file_log(enabled=True, level="INFO"):
    """
    Sets whether the log statements are exported to a log file, and,
    if enabled, the log level to use. If not explicitly configured, file
    logging is OFF.

    Args:
        enabled (bool, optional): Whether to enable file logging.
            This argument defaults to True.
        level (str, optional): If `enabled` is True, sets the level that the
            logging module displays. This level is the minimum and all
            messages at a higher level are included. Valid values are

                - CRITICAL
                - ERROR
                - WARNING
                - INFO
                - DEBUG

            Defaults to "INFO".
    """
    global LOG_TO_FILE
    global FILE_LEVEL
    LOG_TO_FILE = enabled
    FILE_LEVEL = level
    _setup_logger()


def _setup_logger():
    """
    Configures the root logger based on the default or user-specified settings.
    As needed, a StreamHandler is created for console logging or FileHandler
    is created for file logging. Either or both are attached to the root
    logger for use throughout FLORIS.

    Returns:
        logging.Logger: The root logger from the `logging` module.
    """
    # Create a logger object for floris
    logger = logging.getLogger(name="floris")
    logger.setLevel(logging.DEBUG)
    # level_styles = {'warning': {'color': 'red', 'bold': False}}
    fmt_console = "%(name)s %(levelname)s %(message)s"
    fmt_file = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"

    file_name = "floris_{:%Y-%m-%d-%H_%M_%S}.log".format(datetime.now())

    # Remove all existing handlers before adding new ones
    for h in logger.handlers.copy():
        logger.removeHandler(h)

    # Configure and add the console handler
    if LOG_TO_CONSOLE:
        console_handler = logging.StreamHandler()
        console_handler.setLevel(CONSOLE_LEVEL)
        console_format = coloredlogs.ColoredFormatter(
            # level_styles=level_styles,
            fmt=fmt_console
        )
        console_handler.setFormatter(console_format)
        console_handler.addFilter(TracebackInfoFilter(clear=True))
        logger.addHandler(console_handler)

    # Configure and add the file handler
    if LOG_TO_FILE:
        file_handler = logging.FileHandler(file_name)
        file_handler.setLevel(FILE_LEVEL)
        file_format = logging.Formatter(fmt_file)
        file_handler.setFormatter(file_format)
        file_handler.addFilter(TracebackInfoFilter(clear=False))
        logger.addHandler(file_handler)

    return logger


class TracebackInfoFilter(logging.Filter):
    """Clear or restore the exception on log records"""

    def __init__(self, clear=True):
        self.clear = clear

    def filter(self, record):
        if self.clear:
            record._stack_info_hidden, record.stack_info = record.stack_info, None
        elif hasattr(record, "_stack_info_hidden"):
            record.stack_info = record._stack_info_hidden
            del record._stack_info_hidden
        return True


class LoggerBase:
    """
    Convenience super-class to any class requiring access to the logging
    module. The virtual property here allows a simple and dynamic method
    for obtaining the correct logger for the calling class.
    """

    @property
    def logger(self):
        return logging.getLogger(
            "{}.{}".format(type(self).__module__, type(self).__name__)
        )



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt


def gaussian(x, mu, sig):
    """
    Compute gaussian function, from https://stackoverflow.com/questions/14873203/plotting-of-1-dimensional-gaussian-distribution-function.

    Args:
        x (np.array): Input variable to Gaussian.
        mu (float): Mean value.
        sig (float): Standard deviation.

    Returns:
        np.array: The resulting Gaussian distribution.
    """
    return np.exp(-np.power(x - mu, 2.0) / (2 * np.power(sig, 2.0)))


def _convert_to_numpy_array(series):
    """
    Convert an input series to NumPy array. Currently, this function
    checks if an object has a `values` attribute and returns that if it does.
    Otherwise, it returns the given input if that input is a `np.ndarray`.

    Args:
        series (pd.Series): Series to convert.

    Returns:
        np.array: Converted Series.
    """
    if hasattr(series, "values"):
        return series.values
    elif isinstance(series, np.ndarray):
        return series


def _calculate_bootstrap_iterations(n):
    """
    Calculate number of bootstrap iterations given length.
    # TODO: What are `bootstrap iterations`?

    Args:
        n (int): Number of points.

    Returns:
        int: Number of bootstrap iterations.
    """
    maximum = 10000
    minimum = 2000
    return int(np.round(max(min(n * np.log10(n), maximum), minimum)))


def _calculate_lower_and_upper_bound(
    bootstrap_array, percentiles, central_estimate=None, method="simple_percentile"
):
    """
    Given resultant bootstrap output array, compute lower and upper bound
    of confidence interval.

    Args:
        bootstrap_array (np.array): array of bootrapped results
            percentiles (np.array): percentile values
            central_estimate (float, optional): if not using simple percentile,
            need to provide the central estimated result. Defaults to None.
        method (str, optional): method for computing bounds. Defaults to
            'simple_percentile'.

    Returns:
        float, float:

            -   lower ci bound
            -   upper ci bound
    """
    if method == "simple_percentile":
        upper, lower = np.percentile(bootstrap_array, percentiles)
    else:
        lower, upper = 2 * central_estimate - np.percentile(
            bootstrap_array, percentiles
        )
    return lower, upper


def _get_confidence_bounds(confidence):
    """

    Get the upper and lower confidence bounds given a desired confidence level.

    Args:
        confidence (float): [description]
        # TODO: ^^

    Returns:
        float, float:

            -   upper confidence bound
            -   lower confidence bound
    """
    return [50 + 0.5 * confidence, 50 - 0.5 * confidence]


def energy_ratio(
    ref_pow_base, test_pow_base, ws_base, ref_pow_con, test_pow_con, ws_con
):
    """
    Compute the balanced energy ratio for a single binned wind direction.

    This function is called to compute a single balanced
    energy ratio calculation for a particular wind direction bin. Note
    the reference turbine should not be the turbine implementing
    control, but should be an unaffected nearby turbine, or a synthetic
    power estimate from a measurement. See :cite:`er-fleming2019initial` and
    :cite:`er-fleming2019continued` for more information.

    References:
        .. bibliography:: /source/zrefs.bib
            :style: unsrt
            :filter: docname in docnames
            :keyprefix: er-

    Args:
        ref_pow_base (np.array): Array of baseline reference turbine
            power.
        test_pow_base (np.array): Array of baseline test turbine power.
        ws_base (np.array): Array of wind speeds for basline.
        ref_pow_con (np.array): Array of controlled reference turbine
            power.
        test_pow_con (np.array): Array of controlled test turbine power.
        ws_con (np.array): Array of wind speeds in control.

    Returns:
        tuple: tuple containing:

            -   **ratio_base** (*float*): Baseline energy ratio.
            -   **ratio_con** (*float*): Controlled enery ratio.
            -   **ratio_diff** (*float*): Difference in energy ratios.
            -   **p_change** (*float*): Percent change in energy ratios.
            -   **counts_base** (*float*): Number of points in baseline.
            -   **counts_con** (*float*): Number of points in
                controlled.
            -   **counts_diff** (*float*): Number of points in diff (min
                (baseline,controlled)).
            -   **counts_pchange** (*float*): Number of points in
                pchange (min(baseline,controlled)).
    """

    # First derive the weighting functions by wind speed
    ws_unique_base = np.unique(ws_base)
    ws_unique_con = np.unique(ws_con)
    ws_unique = np.intersect1d(ws_unique_base, ws_unique_con)

    if len(ws_unique) == 0:
        return np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan

    # Mask down to the items in both sides
    base_mask = np.isin(ws_base, ws_unique)
    con_mask = np.isin(ws_con, ws_unique)
    ref_pow_base = ref_pow_base[base_mask]
    test_pow_base = test_pow_base[base_mask]
    ws_base = ws_base[base_mask]
    ref_pow_con = ref_pow_con[con_mask]
    test_pow_con = test_pow_con[con_mask]
    ws_con = ws_con[con_mask]

    ws_unique_base, counts_base = np.unique(ws_base, return_counts=True)
    ws_unique_con, counts_con = np.unique(ws_con, return_counts=True)
    total_counts = counts_base + counts_con

    # Make the weights per wind speed
    weights_base = total_counts.astype(float) / counts_base.astype(float)
    weights_con = total_counts.astype(float) / counts_con.astype(float)

    # Make a weighting array
    lut_base = np.zeros(np.max(ws_unique) + 1)
    lut_base[ws_unique] = weights_base
    weight_array_base = lut_base[ws_base]
    lut_con = np.zeros(np.max(ws_unique) + 1)
    lut_con[ws_unique] = weights_con
    weight_array_con = lut_con[ws_con]

    # Weighted sums
    weight_sum_ref_base = np.sum(ref_pow_base * weight_array_base)
    weight_sum_test_base = np.sum(test_pow_base * weight_array_base)
    weight_sum_ref_con = np.sum(ref_pow_con * weight_array_con)
    weight_sum_test_con = np.sum(test_pow_con * weight_array_con)

    # Ratio and diff
    ratio_base = weight_sum_test_base / weight_sum_ref_base
    ratio_con = weight_sum_test_con / weight_sum_ref_con
    ratio_diff = ratio_con - ratio_base
    p_change = 100.0 * ratio_diff / ratio_base

    # Get the counts
    counts_base = len(ref_pow_base)
    counts_con = len(ref_pow_con)
    counts_diff = np.min([counts_base, counts_con])
    counts_pchange = counts_diff

    return (
        ratio_base,
        ratio_con,
        ratio_diff,
        p_change,
        counts_base,
        counts_con,
        counts_diff,
        counts_pchange,
    )


def calculate_balanced_energy_ratio(
    reference_power_baseline,
    test_power_baseline,
    wind_speed_array_baseline,
    wind_direction_array_baseline,
    reference_power_controlled,
    test_power_controlled,
    wind_speed_array_controlled,
    wind_direction_array_controlled,
    wind_direction_bins,
    confidence=95,
    n_boostrap=None,
    wind_direction_bin_p_overlap=None,
):
    """
    Calculate a balanced energy ratio for each wind direction bin.

    Calculate a balanced energy ratio for each wind direction bin.  A
    reference and test turbine are provided for the ratio, as well as
    wind speed and wind directions. These data are further divided into
    baseline and controlled conditions.  The balanced energy ratio
    function is called and used to ensure a similar distribution of
    wind speeds is used in the computation, per wind direction bin, for
    baseline and controlled results.  Resulting arrays, including upper
    and lower uncertaintity bounds computed through bootstrapping, are
    returned.  Note the reference turbine should not be the turbine
    implementing control, but should be an unaffected nearby turbine,
    or a synthetic power estimate from a measurement. See
    :cite:`er-fleming2019initial` and
    :cite:`er-fleming2019continued` for more information.

    References:
        .. bibliography:: /source/zrefs.bib
            :style: unsrt
            :filter: docname in docnames
            :keyprefix: er-

    Args:
        reference_power_baseline (np.array): Array of power of
            reference turbine in baseline conditions.
        test_power_baseline (np.array): Array of power of test turbine
            in baseline conditions.
        wind_speed_array_baseline (np.array): Array of wind speeds in
            baseline conditions.
        wind_direction_array_baseline (np.array): Array of wind
            directions in baseline case.
        reference_power_controlled (np.array): Array of power of
            reference turbine in controlled conditions.
        test_power_controlled (np.array): Array of power of test
            turbine in controlled conditions.
        wind_speed_array_controlled (np.array): Array of wind speeds in
            controlled conditions.
        wind_direction_array_controlled (np.array): Array of wind
            directions in controlled case.
        wind_direction_bins (np.array): Wind directions bins.
        confidence (int, optional): Confidence level to use.  Defaults
            to 95.
        n_boostrap (int, optional): Number of bootstaps, if none,
            _calculate_bootstrap_iterations is called.  Defaults to
            None.
        wind_direction_bin_p_overlap (np.array, optional): Percentage
            overlap between wind direction bin. Defaults to None.

    Returns:
        tuple: tuple containing:

            **ratio_array_base** (*np.array*): Baseline energy ratio at each wind direction bin.
            **lower_ratio_array_base** (*np.array*): Lower confidence bound of baseline energy ratio at each wind direction bin.
            **upper_ratio_array_base** (*np.array*): Upper confidence bound of baseline energy ratio at each wind direction bin.
            **counts_ratio_array_base** (*np.array*): Counts per wind direction bin in baseline.
            **ratio_array_con** (*np.array*): Controlled energy ratio at each wind direction bin.
            **lower_ratio_array_con** (*np.array*): Lower confidence bound of controlled energy ratio at each wind direction bin.
            **upper_ratio_array_con** (*np.array*): Upper confidence bound of controlled energy ratio at each wind direction bin.
            **counts_ratio_array_con** (*np.array*): Counts per wind direction bin in controlled.
            **diff_array** (*np.array*): Difference in baseline and controlled energy ratio per wind direction bin.
            **lower_diff_array** (*np.array*): Lower confidence bound of difference in baseline and controlled energy ratio per wind direction bin.
            **upper_diff_array** (*np.array*): Upper confidence bound of difference in baseline and controlled energy ratio per wind direction bin.
            **counts_diff_array** (*np.array*): Counts in difference (minimum of baseline and controlled).
            **p_change_array** (*np.array*): Percent change in baseline and controlled energy ratio per wind direction bin.
            **lower_p_change_array** (*np.array*): Lower confidence bound of percent change in baseline and controlled energy ratio per wind direction bin.
            **upper_p_change_array** (*np.array*): Upper confidence bound of percent change in baseline and controlled energy ratio per wind direction bin.
            **counts_p_change_array** (*np.array*): Counts in percent change bins (minimum of baseline and controlled).

    """

    # Ensure that input arrays are np.ndarray
    reference_power_baseline = _convert_to_numpy_array(reference_power_baseline)
    test_power_baseline = _convert_to_numpy_array(test_power_baseline)
    wind_speed_array_baseline = _convert_to_numpy_array(wind_speed_array_baseline)
    wind_direction_array_baseline = _convert_to_numpy_array(
        wind_direction_array_baseline
    )

    reference_power_controlled = _convert_to_numpy_array(reference_power_controlled)
    test_power_controlled = _convert_to_numpy_array(test_power_controlled)
    wind_speed_array_controlled = _convert_to_numpy_array(wind_speed_array_controlled)
    wind_direction_array_controlled = _convert_to_numpy_array(
        wind_direction_array_controlled
    )

    # Handle no overlap specificed (assume non-overlap)
    if wind_direction_bin_p_overlap is None:
        wind_direction_bin_p_overlap = 0

    # Compute binning radius (is this right?)
    wind_direction_bin_radius = (
        (1.0 + wind_direction_bin_p_overlap / 100.0)
        * (wind_direction_bins[1] - wind_direction_bins[0])
        / 2.0
    )

    ratio_array_base = np.zeros(len(wind_direction_bins)) * np.nan
    lower_ratio_array_base = np.zeros(len(wind_direction_bins)) * np.nan
    upper_ratio_array_base = np.zeros(len(wind_direction_bins)) * np.nan
    counts_ratio_array_base = np.zeros(len(wind_direction_bins)) * np.nan

    ratio_array_con = np.zeros(len(wind_direction_bins)) * np.nan
    lower_ratio_array_con = np.zeros(len(wind_direction_bins)) * np.nan
    upper_ratio_array_con = np.zeros(len(wind_direction_bins)) * np.nan
    counts_ratio_array_con = np.zeros(len(wind_direction_bins)) * np.nan

    diff_array = np.zeros(len(wind_direction_bins)) * np.nan
    lower_diff_array = np.zeros(len(wind_direction_bins)) * np.nan
    upper_diff_array = np.zeros(len(wind_direction_bins)) * np.nan
    counts_diff_array = np.zeros(len(wind_direction_bins)) * np.nan

    p_change_array = np.zeros(len(wind_direction_bins)) * np.nan
    lower_p_change_array = np.zeros(len(wind_direction_bins)) * np.nan
    upper_p_change_array = np.zeros(len(wind_direction_bins)) * np.nan
    counts_p_change_array = np.zeros(len(wind_direction_bins)) * np.nan

    for i, wind_direction_bin in enumerate(wind_direction_bins):

        wind_dir_mask_baseline = (
            wind_direction_array_baseline
            >= wind_direction_bin - wind_direction_bin_radius
        ) & (
            wind_direction_array_baseline
            < wind_direction_bin + wind_direction_bin_radius
        )

        wind_dir_mask_controlled = (
            wind_direction_array_controlled
            >= wind_direction_bin - wind_direction_bin_radius
        ) & (
            wind_direction_array_controlled
            < wind_direction_bin + wind_direction_bin_radius
        )

        reference_power_baseline_wd = reference_power_baseline[wind_dir_mask_baseline]
        test_power_baseline_wd = test_power_baseline[wind_dir_mask_baseline]
        wind_speed_array_baseline_wd = wind_speed_array_baseline[wind_dir_mask_baseline]
        wind_dir_array_baseline_wd = wind_direction_array_baseline[
            wind_dir_mask_baseline
        ]
        # wd_baseline_dist = (wind_dir_array_baseline_wd - wind_direction_bin)**2
        # # if wind_direction_bin_p_overlap > 1:

        # else:
        #     baseline_weight = np.ones_like(reference_power_baseline_wd)
        # if wind_direction_bin_radius > 1.5:
        if wind_direction_bin_p_overlap > 5.0:
            baseline_weight = gaussian(
                wind_dir_array_baseline_wd,
                wind_direction_bin,
                wind_direction_bin_radius / 2.0,
            )
        else:
            baseline_weight = np.ones_like(wind_dir_array_baseline_wd)
        baseline_weight = baseline_weight / np.sum(baseline_weight)

        reference_power_controlled_wd = reference_power_controlled[
            wind_dir_mask_controlled
        ]
        test_power_controlled_wd = test_power_controlled[wind_dir_mask_controlled]
        wind_speed_array_controlled_wd = wind_speed_array_controlled[
            wind_dir_mask_controlled
        ]
        wind_dir_array_controlled_wd = wind_direction_array_controlled[
            wind_dir_mask_controlled
        ]
        # if wind_direction_bin_radius > 1.5:
        if wind_direction_bin_p_overlap > 5.0:
            controlled_weight = gaussian(
                wind_dir_array_controlled_wd,
                wind_direction_bin,
                wind_direction_bin_radius / 2.0,
            )
        else:
            controlled_weight = np.ones_like(wind_dir_array_controlled_wd)

        controlled_weight = controlled_weight / np.sum(controlled_weight)
        # wd_controlled_dist = (wind_dir_array_controlled_wd - wind_direction_bin)**2

        if (len(reference_power_baseline_wd) == 0) or (
            len(reference_power_controlled_wd) == 0
        ):
            continue

        # Convert wind speed to integers
        wind_speed_array_baseline_wd = wind_speed_array_baseline_wd.round().astype(int)
        wind_speed_array_controlled_wd = wind_speed_array_controlled_wd.round().astype(
            int
        )

        # compute the energy ratio
        (
            ratio_array_base[i],
            ratio_array_con[i],
            diff_array[i],
            p_change_array[i],
            counts_ratio_array_base[i],
            counts_ratio_array_con[i],
            counts_diff_array[i],
            counts_p_change_array[i],
        ) = energy_ratio(
            reference_power_baseline_wd,
            test_power_baseline_wd,
            wind_speed_array_baseline_wd,
            reference_power_controlled_wd,
            test_power_controlled_wd,
            wind_speed_array_controlled_wd,
        )

        # Get the bounds through boot strapping
        # determine the number of bootstrap iterations if not given
        if n_boostrap is None:
            n_boostrap = _calculate_bootstrap_iterations(
                len(reference_power_baseline_wd)
            )

        ratio_base_bs = np.zeros(n_boostrap)
        ratio_con_bs = np.zeros(n_boostrap)
        diff_bs = np.zeros(n_boostrap)
        p_change_bs = np.zeros(n_boostrap)
        for i_bs in range(n_boostrap):

            # random resampling w/ replacement
            # ind_bs = np.random.randint(
            #     len(reference_power_baseline_wd), size=len(reference_power_baseline_wd))
            if wind_direction_bin_p_overlap > 1:
                ind_bs = np.random.choice(
                    len(reference_power_baseline_wd),
                    size=len(reference_power_baseline_wd),
                    p=baseline_weight,
                )
            else:
                ind_bs = np.random.choice(
                    len(reference_power_baseline_wd),
                    size=len(reference_power_baseline_wd),
                )
            reference_power_binned_baseline = reference_power_baseline_wd[ind_bs]
            test_power_binned_baseline = test_power_baseline_wd[ind_bs]
            wind_speed_binned_baseline = wind_speed_array_baseline_wd[ind_bs]

            # ind_bs = np.random.randint(
            #     len(reference_power_controlled_wd), size=len(reference_power_controlled_wd))
            if wind_direction_bin_p_overlap > 1:
                ind_bs = np.random.choice(
                    len(reference_power_controlled_wd),
                    size=len(reference_power_controlled_wd),
                    p=controlled_weight,
                )
            else:
                ind_bs = np.random.choice(
                    len(reference_power_controlled_wd),
                    size=len(reference_power_controlled_wd),
                )
            reference_power_binned_controlled = reference_power_controlled_wd[ind_bs]
            test_power_binned_controlled = test_power_controlled_wd[ind_bs]
            wind_speed_binned_controlled = wind_speed_array_controlled_wd[ind_bs]

            # compute the energy ratio
            (
                ratio_base_bs[i_bs],
                ratio_con_bs[i_bs],
                diff_bs[i_bs],
                p_change_bs[i_bs],
                _,
                _,
                _,
                _,
            ) = energy_ratio(
                reference_power_binned_baseline,
                test_power_binned_baseline,
                wind_speed_binned_baseline,
                reference_power_binned_controlled,
                test_power_binned_controlled,
                wind_speed_binned_controlled,
            )

        # Get the confidence bounds
        percentiles = _get_confidence_bounds(confidence)

        # Compute the central over from the bootstrap runs
        # print('wd',wind_direction_bin)
        # print(ratio_base_bs)
        # print(np.mean(ratio_base_bs))

        # Trim out the nans
        # print('size before',len(ratio_base_bs))
        is_not_nan = (
            (~np.isnan(ratio_base_bs))
            & (~np.isnan(ratio_con_bs))
            & (~np.isnan(diff_bs))
            & (~np.isnan(p_change_bs))
        )
        ratio_base_bs = ratio_base_bs[is_not_nan]
        ratio_con_bs = ratio_con_bs[is_not_nan]
        diff_bs = diff_bs[is_not_nan]
        p_change_bs = p_change_bs[is_not_nan]

        # print('size after',len(ratio_base_bs))

        if len(ratio_base_bs) == 0:
            ratio_array_base[i] = np.nan
            ratio_array_con[i] = np.nan
            diff_array[i] = np.nan
            p_change_array[i] = np.nan
            lower_ratio_array_base[i] = np.nan
            upper_ratio_array_base[i] = np.nan
            lower_ratio_array_con[i] = np.nan
            upper_ratio_array_con[i] = np.nan
            lower_diff_array[i] = np.nan
            upper_diff_array[i] = np.nan
            lower_p_change_array[i] = np.nan
            upper_p_change_array[i] = np.nan
        else:
            ratio_array_base[i] = np.mean(ratio_base_bs)
            ratio_array_con[i] = np.mean(ratio_con_bs)
            diff_array[i] = np.mean(diff_bs)
            p_change_array[i] = np.mean(p_change_bs)

            (
                lower_ratio_array_base[i],
                upper_ratio_array_base[i],
            ) = _calculate_lower_and_upper_bound(
                ratio_base_bs,
                percentiles,
                central_estimate=ratio_array_base[i],
                method="simple_percentile",
            )
            (
                lower_ratio_array_con[i],
                upper_ratio_array_con[i],
            ) = _calculate_lower_and_upper_bound(
                ratio_con_bs,
                percentiles,
                central_estimate=ratio_array_con[i],
                method="simple_percentile",
            )
            lower_diff_array[i], upper_diff_array[i] = _calculate_lower_and_upper_bound(
                diff_bs,
                percentiles,
                central_estimate=diff_array[i],
                method="simple_percentile",
            )
            (
                lower_p_change_array[i],
                upper_p_change_array[i],
            ) = _calculate_lower_and_upper_bound(
                p_change_bs,
                percentiles,
                central_estimate=p_change_array[i],
                method="simple_percentile",
            )

    return (
        ratio_array_base,
        lower_ratio_array_base,
        upper_ratio_array_base,
        counts_ratio_array_base,
        ratio_array_con,
        lower_ratio_array_con,
        upper_ratio_array_con,
        counts_ratio_array_con,
        diff_array,
        lower_diff_array,
        upper_diff_array,
        counts_diff_array,
        p_change_array,
        lower_p_change_array,
        upper_p_change_array,
        counts_p_change_array,
    )


def plot_energy_ratio(
    reference_power_baseline,
    test_power_baseline,
    wind_speed_array_baseline,
    wind_direction_array_baseline,
    reference_power_controlled,
    test_power_controlled,
    wind_speed_array_controlled,
    wind_direction_array_controlled,
    wind_direction_bins,
    confidence=95,
    n_boostrap=None,
    wind_direction_bin_p_overlap=None,
    axarr=None,
    base_color="b",
    con_color="g",
    label_array=None,
    label_pchange=None,
    plot_simple=False,
    plot_ratio_scatter=False,
    marker_scale=1.0,
    show_count=True,
    hide_controlled_case=False,
    ls="--",
    marker=None,
):
    """
    Plot the balanced energy ratio.

    Function mainly acts as a wrapper to call
    calculate_balanced_energy_ratio and plot the results.

    Args:
        reference_power_baseline (np.array): Array of power
            of reference turbine in baseline conditions.
        test_power_baseline (np.array): Array of power of
            test turbine in baseline conditions.
        wind_speed_array_baseline (np.array): Array of wind
            speeds in baseline conditions.
        wind_direction_array_baseline (np.array): Array of
            wind directions in baseline case.
        reference_power_controlled (np.array): Array of power
            of reference turbine in controlled conditions.
        test_power_controlled (np.array): Array of power of
            test turbine in controlled conditions.
        wind_speed_array_controlled (np.array): Array of wind
            speeds in controlled conditions.
        wind_direction_array_controlled (np.array): Array of
            wind directions in controlled case.
        wind_direction_bins (np.array): Wind directions bins.
        confidence (int, optional): Confidence level to use.
            Defaults to 95.
        n_boostrap (int, optional): Number of bootstaps, if
            none, _calculate_bootstrap_iterations is called.  Defaults
            to None.
        wind_direction_bin_p_overlap (np.array, optional):
            Percentage overlap between wind direction bin. Defaults to
            None.
        axarr ([axes], optional): list of axes to plot to.
            Defaults to None.
        base_color (str, optional): Color of baseline in
            plots. Defaults to 'b'.
        con_color (str, optional): Color of controlled in
            plots. Defaults to 'g'.
        label_array ([str], optional): List of labels to
            apply Defaults to None.
        label_pchange ([type], optional): Label for
            percentage change. Defaults to None.
        plot_simple (bool, optional): Plot only the ratio, no
            confidence. Defaults to False.
        plot_ratio_scatter (bool, optional): Include scatter
            plot of values, sized to indicate counts. Defaults to False.
        marker_scale ([type], optional): Marker scale.
            Defaults to 1.
        show_count (bool, optional): Show the counts as scatter plot
        hide_controlled_case (bool, optional): Option to hide the control case from plots, for demonstration

    """

    if axarr is None:
        fig, axarr = plt.subplots(3, 1, sharex=True)

    if label_array is None:
        label_array = ["Baseline", "Controlled"]

    if label_pchange is None:
        label_pchange = "Energy Gain"

    (
        ratio_array_base,
        lower_ratio_array_base,
        upper_ratio_array_base,
        counts_ratio_array_base,
        ratio_array_con,
        lower_ratio_array_con,
        upper_ratio_array_con,
        counts_ratio_array_con,
        diff_array,
        lower_diff_array,
        upper_diff_array,
        counts_diff_array,
        p_change_array,
        lower_p_change_array,
        upper_p_change_array,
        counts_p_change_array,
    ) = calculate_balanced_energy_ratio(
        reference_power_baseline,
        test_power_baseline,
        wind_speed_array_baseline,
        wind_direction_array_baseline,
        reference_power_controlled,
        test_power_controlled,
        wind_speed_array_controlled,
        wind_direction_array_controlled,
        wind_direction_bins,
        confidence=95,
        n_boostrap=n_boostrap,
        wind_direction_bin_p_overlap=wind_direction_bin_p_overlap,
    )

    if plot_simple:
        ax = axarr[0]
        ax.plot(
            wind_direction_bins,
            ratio_array_base,
            label=label_array[0],
            color=base_color,
            ls=ls,
            marker=marker,
        )
        if not hide_controlled_case:
            ax.plot(
                wind_direction_bins,
                ratio_array_con,
                label=label_array[1],
                color=con_color,
                ls=ls,
                marker=marker,
            )
        ax.axhline(1, color="k")
        ax.set_ylabel("Energy Ratio (-)")

        ax = axarr[1]
        ax.plot(
            wind_direction_bins,
            diff_array,
            label=label_pchange,
            color=con_color,
            ls=ls,
            marker=marker,
        )
        ax.axhline(0, color="k")
        ax.set_ylabel("Change in Energy Ratio (-)")

        ax = axarr[2]
        ax.plot(
            wind_direction_bins,
            p_change_array,
            label=label_pchange,
            color=con_color,
            ls=ls,
            marker=marker,
        )
        ax.axhline(0, color="k")
        ax.set_ylabel("% Change in Energy Ratio (-)")

    else:

        ax = axarr[0]
        ax.plot(
            wind_direction_bins,
            ratio_array_base,
            label=label_array[0],
            color=base_color,
            ls="-",
            marker=".",
        )
        ax.fill_between(
            wind_direction_bins,
            lower_ratio_array_base,
            upper_ratio_array_base,
            alpha=0.3,
            color=base_color,
            label="_nolegend_",
        )
        if show_count:
            ax.scatter(
                wind_direction_bins,
                ratio_array_base,
                s=counts_ratio_array_base * marker_scale,
                label="_nolegend_",
                color=base_color,
                marker="o",
                alpha=0.2,
            )
        if not hide_controlled_case:
            ax.plot(
                wind_direction_bins,
                ratio_array_con,
                label=label_array[1],
                color=con_color,
                ls="-",
                marker=".",
            )
            ax.fill_between(
                wind_direction_bins,
                lower_ratio_array_con,
                upper_ratio_array_con,
                alpha=0.3,
                color=con_color,
                label="_nolegend_",
            )
            if show_count:
                ax.scatter(
                    wind_direction_bins,
                    ratio_array_con,
                    s=counts_ratio_array_con * marker_scale,
                    label="_nolegend_",
                    color=con_color,
                    marker="o",
                    alpha=0.2,
                )
        ax.axhline(1, color="k")
        ax.set_ylabel("Energy Ratio (-)")

        ax = axarr[1]
        ax.plot(
            wind_direction_bins,
            diff_array,
            label=label_pchange,
            color=con_color,
            ls="-",
            marker=".",
        )
        ax.fill_between(
            wind_direction_bins,
            lower_diff_array,
            upper_diff_array,
            alpha=0.3,
            color=con_color,
            label="_nolegend_",
        )
        if show_count:
            ax.scatter(
                wind_direction_bins,
                diff_array,
                s=counts_diff_array * marker_scale,
                label="_nolegend_",
                color=con_color,
                marker="o",
                alpha=0.2,
            )
        ax.axhline(0, color="k")
        ax.set_ylabel("Change in Energy Ratio (-)")

        ax = axarr[2]
        ax.plot(
            wind_direction_bins,
            p_change_array,
            label=label_pchange,
            color=con_color,
            ls="-",
            marker=".",
        )
        ax.fill_between(
            wind_direction_bins,
            lower_p_change_array,
            upper_p_change_array,
            alpha=0.3,
            color=con_color,
            label="_nolegend_",
        )
        if show_count:
            ax.scatter(
                wind_direction_bins,
                p_change_array,
                s=counts_p_change_array * marker_scale,
                label="_nolegend_",
                color=con_color,
                marker="o",
                alpha=0.2,
            )
        ax.axhline(0, color="k")
        ax.set_ylabel("% Change in Energy Ratio (-)")

    for ax in axarr:
        ax.grid(True)
        ax.set_xlabel("Wind Direction (Deg)")

    return diff_array


# ######WIND SPEED VERSIONS##################


def energy_ratio_ws(
    ref_pow_base,
    test_pow_base,
    wd_base,
    ref_pow_con,
    test_pow_con,
    wd_con,
    use_absolutes=False,
    use_mean=False,
):
    """
    Compute the balanced energy ratio, however, inverted so balancing against wind direction, compute for a single wind speed bin

    This function is typically called to compute a single balanced
    energy ratio calculation for a particular wind speed bin.  Note
    the reference turbine should not be the turbine implementing
    control, but should be an unaffected nearby turbine, or a synthetic
    power estimate from a measurement.

    Args:
        ref_pow_base (np.array): Array of baseline reference turbine
            power.
        test_pow_base (np.array): Array of baseline test turbine power.
        wd_base (np.array): Array of wind directions for basline.
        ref_pow_con (np.array): Array of controlled reference turbine
            power.
        test_pow_con (np.array): Array of controlled test turbine power.
        wd_con (np.array): Array of wind directions in control.
        use_absolutes (boolean): Should use ratios or absolute energy values
        use_mean (boolean): If using absolutes, use the mean (power), or the sum (energy)


    Returns:
        tuple: tuple containing:

            -   **ratio_base** (*float*): Baseline energy ratio.
            -   **ratio_con** (*float*): Controlled enery ra tio.
            -   **ratio_diff** (*float*): Difference in energy ratios.
            -   **p_change** (*float*): Percent change in energy ratios.
            -   **counts_base** (*float*): Number of points in baseline.
            -   **counts_con** (*float*): Number of points in
                controlled.
            -   **counts_diff** (*float*): Number of points in diff (min
                (baseline,controlled)).
            -   **counts_pchange** (*float*): Number of points in
                pchange (min(baseline,controlled)).
    """

    # First derive the weighting functions by wind direction
    wd_unique_base = np.unique(wd_base)
    wd_unique_con = np.unique(wd_con)
    wd_unique = np.intersect1d(wd_unique_base, wd_unique_con)

    if len(wd_unique) == 0:
        return np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan

    # Mask down to the items in both sides
    base_mask = np.isin(wd_base, wd_unique)
    con_mask = np.isin(wd_con, wd_unique)
    ref_pow_base = ref_pow_base[base_mask]
    test_pow_base = test_pow_base[base_mask]
    wd_base = wd_base[base_mask]
    ref_pow_con = ref_pow_con[con_mask]
    test_pow_con = test_pow_con[con_mask]
    wd_con = wd_con[con_mask]

    wd_unique_base, counts_base = np.unique(wd_base, return_counts=True)
    wd_unique_con, counts_con = np.unique(wd_con, return_counts=True)
    total_counts = counts_base + counts_con

    # Make the weights per wind speed
    weights_base = total_counts.astype(float) / counts_base.astype(float)
    weights_con = total_counts.astype(float) / counts_con.astype(float)

    # Make a weighting array
    lut_base = np.zeros(np.max(wd_unique) + 1)
    lut_base[wd_unique] = weights_base
    weight_array_base = lut_base[wd_base]
    lut_con = np.zeros(np.max(wd_unique) + 1)
    lut_con[wd_unique] = weights_con
    weight_array_con = lut_con[wd_con]

    # Weighted sums
    weight_sum_ref_base = 2 * np.sum(ref_pow_base * weight_array_base)
    weight_sum_test_base = 2 * np.sum(test_pow_base * weight_array_base)
    weight_sum_ref_con = 2 * np.sum(ref_pow_con * weight_array_con)
    weight_sum_test_con = 2 * np.sum(test_pow_con * weight_array_con)

    # Ratio and diff
    if use_absolutes:
        ratio_base = weight_sum_test_base
        ratio_con = weight_sum_test_con
        if use_mean:
            ratio_base = np.mean(test_pow_base)
            ratio_con = np.mean(test_pow_con)
    else:

        ratio_base = weight_sum_test_base / weight_sum_ref_base
        ratio_con = weight_sum_test_con / weight_sum_ref_con

    ratio_diff = ratio_con - ratio_base
    p_change = 100.0 * ratio_diff / ratio_base

    # Get the counts
    counts_base = len(ref_pow_base)
    counts_con = len(ref_pow_con)
    counts_diff = np.min([counts_base, counts_con])
    counts_pchange = counts_diff

    return (
        ratio_base,
        ratio_con,
        ratio_diff,
        p_change,
        counts_base,
        counts_con,
        counts_diff,
        counts_pchange,
    )


def calculate_balanced_energy_ratio_ws(
    reference_power_baseline,
    test_power_baseline,
    wind_speed_array_baseline,
    wind_direction_array_baseline,
    reference_power_controlled,
    test_power_controlled,
    wind_speed_array_controlled,
    wind_direction_array_controlled,
    wind_speed_bins,
    confidence=95,
    n_boostrap=None,
    wind_speed_bin_p_overlap=None,
    use_absolutes=False,
    use_mean=False,
):
    """
    Calculate a balanced energy ratio for each wind speed bin.

    Calculate a balanced energy ratio for each wind speed bin.  A
    reference and test turbine are provided for the ratio, as well as
    wind speed and wind directions. These data are further divided into
    baseline and controlled conditions.  The balanced energy ratio
    function is called and used to ensure a similar distribution of
    wind speeds is used in the computation, per wind speed bin, for
    baseline and controlled results.  Resulting arrays, including upper
    and lower uncertainty bounds computed through bootstrapping, are
    returned.  Note the reference turbine should not be the turbine
    implementing control, but should be an unaffected nearby turbine,
    or a synthetic power estimate from a measurement

    Args:
        reference_power_baseline (np.array): Array of power of
            reference turbine in baseline conditions.
        test_power_baseline (np.array): Array of power of test turbine
            in baseline conditions.
        wind_speed_array_baseline (np.array): Array of wind speeds in
            baseline conditions.
        wind_direction_array_baseline (np.array): Array of wind
            directions in baseline case.
        reference_power_controlled (np.array): Array of power of
            reference turbine in controlled conditions.
        test_power_controlled (np.array): Array of power of test
            turbine in controlled conditions.
        wind_speed_array_controlled (np.array): Array of wind speeds in
            controlled conditions.
        wind_direction_array_controlled (np.array): Array of wind
            directions in controlled case.
        wind_speed_bins (np.array): Wind speed bins.
        confidence (int, optional): Confidence level to use.  Defaults
            to 95.
        n_boostrap (int, optional): Number of bootstaps, if none,
            _calculate_bootstrap_iterations is called.  Defaults to
            None.
        wind_speed_bin_p_overlap (np.array, optional): Percentage
            overlap between wind direction bin. Defaults to None.
        use_absolutes (boolean): Should use ratios or absolute energy values
        use_mean (boolean): If using absolutes, use the mean (power), or the sum (energy)

    Returns:
        tuple: tuple containing:

            **ratio_array_base** (*np.array*): Baseline energy ratio at each wind speed bin.
            **lower_ratio_array_base** (*np.array*): Lower confidence bound of baseline energy ratio at each wind speed bin.
            **upper_ratio_array_base** (*np.array*): Upper confidence bound of baseline energy ratio at each wind speed bin.
            **counts_ratio_array_base** (*np.array*): Counts per wind speed bin in baseline.
            **ratio_array_con** (*np.array*): Controlled energy ratio at each wind speed bin.
            **lower_ratio_array_con** (*np.array*): Lower confidence bound of controlled energy ratio at each wind speed bin.
            **upper_ratio_array_con** (*np.array*): Upper confidence bound of controlled energy ratio at each wind speed bin.
            **counts_ratio_array_con** (*np.array*): Counts per wind speed bin in controlled.
            **diff_array** (*np.array*): Difference in baseline and controlled energy ratio per wind speed bin.
            **lower_diff_array** (*np.array*): Lower confidence bound of difference in baseline and controlled energy ratio per wind speed bin.
            **upper_diff_array** (*np.array*): Upper confidence bound of difference in baseline and controlled energy ratio per wind speed bin.
            **counts_diff_array** (*np.array*): Counts in difference (minimum of baseline and controlled).
            **p_change_array** (*np.array*): Percent change in baseline and controlled energy ratio per wind speed bin.
            **lower_p_change_array** (*np.array*): Lower confidence bound of percent change in baseline and controlled energy ratio per wind speed bin.
            **upper_p_change_array** (*np.array*): Upper confidence bound of percent change in baseline and controlled energy ratio per wind speed bin.
            **counts_p_change_array** (*np.array*): Counts in percent change bins (minimum of baseline and controlled).

    """

    # Ensure that input arrays are np.ndarray
    reference_power_baseline = _convert_to_numpy_array(reference_power_baseline)
    test_power_baseline = _convert_to_numpy_array(test_power_baseline)
    wind_speed_array_baseline = _convert_to_numpy_array(wind_speed_array_baseline)
    wind_direction_array_baseline = _convert_to_numpy_array(
        wind_direction_array_baseline
    )

    reference_power_controlled = _convert_to_numpy_array(reference_power_controlled)
    test_power_controlled = _convert_to_numpy_array(test_power_controlled)
    wind_speed_array_controlled = _convert_to_numpy_array(wind_speed_array_controlled)
    wind_direction_array_controlled = _convert_to_numpy_array(
        wind_direction_array_controlled
    )

    # Handle no overlap specificed (assume non-overlap)
    if wind_speed_bin_p_overlap is None:
        wind_speed_bin_p_overlap = 0

    # Compute binning radius (is this right?)
    wind_speed_bin_radius = (
        (1.0 + wind_speed_bin_p_overlap / 100.0)
        * (wind_speed_bins[1] - wind_speed_bins[0])
        / 2.0
    )

    ratio_array_base = np.zeros(len(wind_speed_bins)) * np.nan
    lower_ratio_array_base = np.zeros(len(wind_speed_bins)) * np.nan
    upper_ratio_array_base = np.zeros(len(wind_speed_bins)) * np.nan
    counts_ratio_array_base = np.zeros(len(wind_speed_bins)) * np.nan

    ratio_array_con = np.zeros(len(wind_speed_bins)) * np.nan
    lower_ratio_array_con = np.zeros(len(wind_speed_bins)) * np.nan
    upper_ratio_array_con = np.zeros(len(wind_speed_bins)) * np.nan
    counts_ratio_array_con = np.zeros(len(wind_speed_bins)) * np.nan

    diff_array = np.zeros(len(wind_speed_bins)) * np.nan
    lower_diff_array = np.zeros(len(wind_speed_bins)) * np.nan
    upper_diff_array = np.zeros(len(wind_speed_bins)) * np.nan
    counts_diff_array = np.zeros(len(wind_speed_bins)) * np.nan

    p_change_array = np.zeros(len(wind_speed_bins)) * np.nan
    lower_p_change_array = np.zeros(len(wind_speed_bins)) * np.nan
    upper_p_change_array = np.zeros(len(wind_speed_bins)) * np.nan
    counts_p_change_array = np.zeros(len(wind_speed_bins)) * np.nan

    for i, wind_speed_bin in enumerate(wind_speed_bins):

        wind_speed_mask_baseline = (
            wind_speed_array_baseline >= wind_speed_bin - wind_speed_bin_radius
        ) & (wind_speed_array_baseline < wind_speed_bin + wind_speed_bin_radius)

        wind_speed_mask_controlled = (
            wind_speed_array_controlled >= wind_speed_bin - wind_speed_bin_radius
        ) & (wind_speed_array_controlled < wind_speed_bin + wind_speed_bin_radius)

        reference_power_baseline_ws = reference_power_baseline[wind_speed_mask_baseline]
        test_power_baseline_ws = test_power_baseline[wind_speed_mask_baseline]
        wind_dir_array_baseline_ws = wind_direction_array_baseline[
            wind_speed_mask_baseline
        ]

        reference_power_controlled_ws = reference_power_controlled[
            wind_speed_mask_controlled
        ]
        test_power_controlled_ws = test_power_controlled[wind_speed_mask_controlled]
        wind_dir_array_controlled_ws = wind_direction_array_controlled[
            wind_speed_mask_controlled
        ]

        if (len(reference_power_baseline_ws) == 0) or (
            len(reference_power_controlled_ws) == 0
        ):
            continue

        # Convert wind speeds to integers
        wind_dir_array_baseline_ws = wind_dir_array_baseline_ws.round().astype(int)
        wind_dir_array_controlled_ws = wind_dir_array_controlled_ws.round().astype(int)

        # compute the energy ratio
        (
            ratio_array_base[i],
            ratio_array_con[i],
            diff_array[i],
            p_change_array[i],
            counts_ratio_array_base[i],
            counts_ratio_array_con[i],
            counts_diff_array[i],
            counts_p_change_array[i],
        ) = energy_ratio_ws(
            reference_power_baseline_ws,
            test_power_baseline_ws,
            wind_dir_array_baseline_ws,
            reference_power_controlled_ws,
            test_power_controlled_ws,
            wind_dir_array_controlled_ws,
            use_absolutes=use_absolutes,
            use_mean=use_mean,
        )

        # Get the bounds through boot strapping
        # determine the number of bootstrap iterations if not given
        if n_boostrap is None:
            n_boostrap = _calculate_bootstrap_iterations(
                len(reference_power_baseline_ws)
            )

        ratio_base_bs = np.zeros(n_boostrap)
        ratio_con_bs = np.zeros(n_boostrap)
        diff_bs = np.zeros(n_boostrap)
        p_change_bs = np.zeros(n_boostrap)
        for i_bs in range(n_boostrap):

            # random resampling w/ replacement
            ind_bs = np.random.randint(
                len(reference_power_baseline_ws), size=len(reference_power_baseline_ws)
            )
            reference_power_binned_baseline = reference_power_baseline_ws[ind_bs]
            test_power_binned_baseline = test_power_baseline_ws[ind_bs]
            wind_dir_binned_baseline = wind_dir_array_baseline_ws[ind_bs]

            ind_bs = np.random.randint(
                len(reference_power_controlled_ws),
                size=len(reference_power_controlled_ws),
            )
            reference_power_binned_controlled = reference_power_controlled_ws[ind_bs]
            test_power_binned_controlled = test_power_controlled_ws[ind_bs]
            wind_dir_binned_controlled = wind_dir_array_controlled_ws[ind_bs]

            # compute the energy ratio
            (
                ratio_base_bs[i_bs],
                ratio_con_bs[i_bs],
                diff_bs[i_bs],
                p_change_bs[i_bs],
                _,
                _,
                _,
                _,
            ) = energy_ratio_ws(
                reference_power_binned_baseline,
                test_power_binned_baseline,
                wind_dir_binned_baseline,
                reference_power_binned_controlled,
                test_power_binned_controlled,
                wind_dir_binned_controlled,
                use_absolutes=use_absolutes,
                use_mean=use_mean,
            )

        # Get the confidence bounds
        percentiles = _get_confidence_bounds(confidence)

        (
            lower_ratio_array_base[i],
            upper_ratio_array_base[i],
        ) = _calculate_lower_and_upper_bound(
            ratio_base_bs,
            percentiles,
            central_estimate=ratio_array_base[i],
            method="simple_percentile",
        )
        (
            lower_ratio_array_con[i],
            upper_ratio_array_con[i],
        ) = _calculate_lower_and_upper_bound(
            ratio_con_bs,
            percentiles,
            central_estimate=ratio_array_con[i],
            method="simple_percentile",
        )
        lower_diff_array[i], upper_diff_array[i] = _calculate_lower_and_upper_bound(
            diff_bs,
            percentiles,
            central_estimate=diff_array[i],
            method="simple_percentile",
        )
        (
            lower_p_change_array[i],
            upper_p_change_array[i],
        ) = _calculate_lower_and_upper_bound(
            p_change_bs,
            percentiles,
            central_estimate=p_change_array[i],
            method="simple_percentile",
        )

    return (
        ratio_array_base,
        lower_ratio_array_base,
        upper_ratio_array_base,
        counts_ratio_array_base,
        ratio_array_con,
        lower_ratio_array_con,
        upper_ratio_array_con,
        counts_ratio_array_con,
        diff_array,
        lower_diff_array,
        upper_diff_array,
        counts_diff_array,
        p_change_array,
        lower_p_change_array,
        upper_p_change_array,
        counts_p_change_array,
    )


def plot_energy_ratio_ws(
    reference_power_baseline,
    test_power_baseline,
    wind_speed_array_baseline,
    wind_direction_array_baseline,
    reference_power_controlled,
    test_power_controlled,
    wind_speed_array_controlled,
    wind_direction_array_controlled,
    wind_speed_bins,
    confidence=95,
    n_boostrap=None,
    wind_speed_bin_p_overlap=None,
    axarr=None,
    base_color="b",
    con_color="g",
    label_array=None,
    label_pchange=None,
    plot_simple=False,
    plot_ratio_scatter=False,
    marker_scale=1.0,
    use_absolutes=False,
    show_power=False,
    show_percent_change=True,
):
    """
    Plot the balanced energy ratio against wind speed

    Function mainly acts as a wrapper to call
    calculate_balanced_energy_ratio and plot the results.

    Args:
        reference_power_baseline (np.array): Array of power
            of reference turbine in baseline conditions.
        test_power_baseline (np.array): Array of power of
            test turbine in baseline conditions.
        wind_speed_array_baseline (np.array): Array of wind
            speeds in baseline conditions.
        wind_direction_array_baseline (np.array): Array of
            wind directions in baseline case.
        reference_power_controlled (np.array): Array of power
            of reference turbine in controlled conditions.
        test_power_controlled (np.array): Array of power of
            test turbine in controlled conditions.
        wind_speed_array_controlled (np.array): Array of wind
            speeds in controlled conditions.
        wind_direction_array_controlled (np.array): Array of
            wind directions in controlled case.
        wind_speed_bins (np.array): Wind speed bins.
        confidence (int, optional): Confidence level to use.
            Defaults to 95.
        n_boostrap (int, optional): Number of bootstaps, if
            none, _calculate_bootstrap_iterations is called.  Defaults
            to None.
        wind_direction_bin_p_overlap (np.array, optional):
            Percentage overlap between wind direction bin. Defaults to
            None.
        axarr ([axes], optional): list of axes to plot to.
            Defaults to None.
        base_color (str, optional): Color of baseline in
            plots. Defaults to 'b'.
        con_color (str, optional): Color of controlled in
            plots. Defaults to 'g'.
        label_array ([str], optional): List of labels to
            apply Defaults to None.
        label_pchange ([type], optional): Label for
            percentage change. Defaults to None.
        plot_simple (bool, optional): Plot only the ratio, no
            confidence. Defaults to False.
        plot_ratio_scatter (bool, optional): Include scatter
            plot of values, sized to indicate counts. Defaults to False.
        marker_scale ([type], optional): Marker scale.
            Defaults to 1.
        use_absolutes (boolean): Should use ratios or absolute energy values
        show_power (boolean): If using absolutes, use the mean (power), or the sum (energy)
        show_percent_change (boolean): Should percent change be shown on 3rd axis
    """

    if axarr is None:
        fig, axarr = plt.subplots(3, 1, sharex=True)

    if label_array is None:
        label_array = ["Baseline", "Controlled"]

    if label_pchange is None:
        label_pchange = "Energy Gain"

    (
        ratio_array_base,
        lower_ratio_array_base,
        upper_ratio_array_base,
        counts_ratio_array_base,
        ratio_array_con,
        lower_ratio_array_con,
        upper_ratio_array_con,
        counts_ratio_array_con,
        diff_array,
        lower_diff_array,
        upper_diff_array,
        counts_diff_array,
        p_change_array,
        lower_p_change_array,
        upper_p_change_array,
        counts_p_change_array,
    ) = calculate_balanced_energy_ratio_ws(
        reference_power_baseline,
        test_power_baseline,
        wind_speed_array_baseline,
        wind_direction_array_baseline,
        reference_power_controlled,
        test_power_controlled,
        wind_speed_array_controlled,
        wind_direction_array_controlled,
        wind_speed_bins,
        confidence=95,
        n_boostrap=None,
        wind_speed_bin_p_overlap=wind_speed_bin_p_overlap,
        use_absolutes=use_absolutes,
        use_mean=show_power,
    )

    # If using absolutes, flip test and reference to to get seperate reference curves
    if use_absolutes:
        (
            r_ratio_array_base,
            r_lower_ratio_array_base,
            r_upper_ratio_array_base,
            r_counts_ratio_array_base,
            r_ratio_array_con,
            r_lower_ratio_array_con,
            r_upper_ratio_array_con,
            r_counts_ratio_array_con,
            r_diff_array,
            r_lower_diff_array,
            r_upper_diff_array,
            r_counts_diff_array,
            r_p_change_array,
            r_lower_p_change_array,
            r_upper_p_change_array,
            r_counts_p_change_array,
        ) = calculate_balanced_energy_ratio_ws(
            test_power_baseline,
            reference_power_baseline,
            wind_speed_array_baseline,
            wind_direction_array_baseline,
            test_power_controlled,
            reference_power_controlled,
            wind_speed_array_controlled,
            wind_direction_array_controlled,
            wind_speed_bins,
            confidence=95,
            n_boostrap=None,
            wind_speed_bin_p_overlap=wind_speed_bin_p_overlap,
            use_absolutes=use_absolutes,
            use_mean=show_power,
        )

    # ENERGY RATIO PLOT
    ax = axarr[0]
    if plot_simple:
        ax.plot(
            wind_speed_bins,
            ratio_array_base,
            label=label_array[0],
            color=base_color,
            ls="--",
        )
        ax.plot(
            wind_speed_bins,
            ratio_array_con,
            label=label_array[1],
            color=con_color,
            ls="--",
        )
    else:
        ax.plot(
            wind_speed_bins,
            ratio_array_base,
            label=label_array[0],
            color=base_color,
            ls="-",
            marker=".",
        )
        ax.fill_between(
            wind_speed_bins,
            lower_ratio_array_base,
            upper_ratio_array_base,
            alpha=0.3,
            color=base_color,
            label="_nolegend_",
        )
        ax.scatter(
            wind_speed_bins,
            ratio_array_base,
            s=counts_ratio_array_base,
            label="_nolegend_",
            color=base_color,
            marker="o",
            alpha=0.2,
        )
        ax.plot(
            wind_speed_bins,
            ratio_array_con,
            label=label_array[1],
            color=con_color,
            ls="-",
            marker=".",
        )
        ax.fill_between(
            wind_speed_bins,
            lower_ratio_array_con,
            upper_ratio_array_con,
            alpha=0.3,
            color=con_color,
            label="_nolegend_",
        )
        ax.scatter(
            wind_speed_bins,
            ratio_array_con,
            s=counts_ratio_array_con,
            label="_nolegend_",
            color=con_color,
            marker="o",
            alpha=0.2,
        )

    if use_absolutes:  # show the references as well
        if plot_simple:
            ax.plot(
                wind_speed_bins,
                r_ratio_array_base,
                label="_nolegend_",
                color="k",
                ls="--",
            )
            ax.plot(
                wind_speed_bins,
                r_ratio_array_con,
                label="_nolegend_",
                color="k",
                ls="--",
            )
        else:
            ax.plot(
                wind_speed_bins,
                r_ratio_array_base,
                label="_nolegend_",
                color="k",
                ls="-",
                marker=".",
            )
            ax.fill_between(
                wind_speed_bins,
                r_lower_ratio_array_base,
                r_upper_ratio_array_base,
                alpha=0.3,
                color="k",
                label="_nolegend_",
            )
            ax.scatter(
                wind_speed_bins,
                r_ratio_array_base,
                s=counts_ratio_array_base,
                label="_nolegend_",
                color="k",
                marker="o",
                alpha=0.2,
            )
            ax.plot(
                wind_speed_bins,
                r_ratio_array_con,
                label="_nolegend_",
                color="k",
                ls="-",
                marker=".",
            )
            ax.fill_between(
                wind_speed_bins,
                r_lower_ratio_array_con,
                r_upper_ratio_array_con,
                alpha=0.3,
                color="k",
                label="_nolegend_",
            )
            ax.scatter(
                wind_speed_bins,
                r_ratio_array_con,
                s=counts_ratio_array_con,
                label="_nolegend_",
                color="k",
                marker="o",
                alpha=0.2,
            )

    if not use_absolutes:
        ax.axhline(1, color="k")
        ax.set_ylabel("Energy Ratio (-)")
    else:
        if show_power:
            ax.set_ylabel("Power (kW)")
        else:
            ax.set_ylabel("Energy (kW-min)")

    # Difference PLOT
    ax = axarr[1]
    if plot_simple:
        ax.plot(
            wind_speed_bins, diff_array, label=label_pchange, color=con_color, ls="--"
        )

    else:
        ax.plot(
            wind_speed_bins,
            diff_array,
            label=label_pchange,
            color=con_color,
            ls="-",
            marker=".",
        )
        ax.fill_between(
            wind_speed_bins,
            lower_diff_array,
            upper_diff_array,
            alpha=0.3,
            color=con_color,
            label="_nolegend_",
        )
        ax.scatter(
            wind_speed_bins,
            diff_array,
            s=counts_diff_array,
            label="_nolegend_",
            color=con_color,
            marker="o",
            alpha=0.2,
        )

    if not use_absolutes:
        ax.axhline(0, color="k")
        ax.set_ylabel("Change in Energy Ratio (-)")
    else:
        if show_power:
            ax.set_ylabel("Change in Power (kW)")
        else:
            ax.set_ylabel("Change in Energy (kW-min)")

    # Percent Change plot PLOT
    if show_percent_change:
        ax = axarr[2]
        if plot_simple:
            ax.plot(
                wind_speed_bins,
                p_change_array,
                label=label_pchange,
                color=con_color,
                ls="--",
            )

        else:
            ax.plot(
                wind_speed_bins,
                p_change_array,
                label=label_pchange,
                color=con_color,
                ls="-",
                marker=".",
            )
            ax.fill_between(
                wind_speed_bins,
                lower_p_change_array,
                upper_p_change_array,
                alpha=0.3,
                color=con_color,
                label="_nolegend_",
            )
            ax.scatter(
                wind_speed_bins,
                p_change_array,
                s=counts_p_change_array,
                label="_nolegend_",
                color=con_color,
                marker="o",
                alpha=0.2,
            )

        if not use_absolutes:
            ax.axhline(0, color="k")
            ax.set_ylabel("%% Change in Energy Ratio (-)")
        else:
            if show_power:
                ax.set_ylabel("%% Change in Power (kW)")
            else:
                ax.set_ylabel("%% Change in Energy (kW-min)")

    for ax in axarr:
        ax.grid(True)
        ax.set_xlabel("Wind Speed (m/s)")



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# See https://floris.readthedocs.io for documentation

from __future__ import annotations

import attrs
from attrs import define, field
import numpy as np

from floris.type_dec import (
    FromDictMixin,
    NDArrayFloat,
    floris_array_converter
)
from floris.simulation import Grid


@define
class FlowField(FromDictMixin):
    wind_speeds: NDArrayFloat = field(converter=floris_array_converter)
    wind_directions: NDArrayFloat = field(converter=floris_array_converter)
    wind_veer: float = field(converter=float)
    wind_shear: float = field(converter=float)
    air_density: float = field(converter=float)
    turbulence_intensity: float = field(converter=float)
    reference_wind_height: int = field(converter=int)

    n_wind_speeds: int = field(init=False)
    n_wind_directions: int = field(init=False)

    u_initial: NDArrayFloat = field(init=False, default=np.array([]))
    v_initial: NDArrayFloat = field(init=False, default=np.array([]))
    w_initial: NDArrayFloat = field(init=False, default=np.array([]))
    u: NDArrayFloat = field(init=False, default=np.array([]))
    v: NDArrayFloat = field(init=False, default=np.array([]))
    w: NDArrayFloat = field(init=False, default=np.array([]))
    het_map: list = field(init=False, default=None)

    turbulence_intensity_field: NDArrayFloat = field(init=False, default=np.array([]))

    @wind_speeds.validator
    def wind_speeds_validator(self, instance: attrs.Attribute, value: NDArrayFloat) -> None:
        """Using the validator method to keep the `n_wind_speeds` attribute up to date."""
        self.n_wind_speeds = value.size

    @wind_directions.validator
    def wind_directions_validator(self, instance: attrs.Attribute, value: NDArrayFloat) -> None:
        """Using the validator method to keep the `n_wind_directions` attribute up to date."""
        self.n_wind_directions = value.size

    def initialize_velocity_field(self, grid: Grid) -> None:

        # Create an initial wind profile as a function of height. The values here will
        # be multiplied with the wind speeds to give the initial wind field.
        # Since we use grid.z, this is a vertical plane for each turbine
        # Here, the profile is of shape (# turbines, N grid points, M grid points)
        # This velocity profile is 1.0 at the reference wind height and then follows wind shear as an exponent.
        # NOTE: the convention of which dimension on the TurbineGrid is vertical and horizontal is
        # determined by this line. Since the right-most dimension on grid.z is storing the values
        # for height, using it here to apply the shear law makes that dimension store the vertical
        # wind profile.
        wind_profile_plane = (grid.z / self.reference_wind_height) ** self.wind_shear

        # If no hetergeneous inflow defined, then set all speeds ups to 1.0
        if self.het_map is None:
            speed_ups = 1.0

        # If heterogeneous flow data is given, the speed ups at the defined
        # grid locations are determined in either 2 or 3 dimensions.
        else:
            if len(self.het_map[0][0].points[0]) == 2:
                speed_ups = self.calculate_speed_ups(self.het_map, grid.x, grid.y)
            elif len(self.het_map[0][0].points[0]) == 3:
                speed_ups = self.calculate_speed_ups(self.het_map, grid.x, grid.y, grid.z)

        # Create the sheer-law wind profile
        # This array is of shape (# wind directions, # wind speeds, grid.template_array)
        # Since generally grid.template_array may be many different shapes, we use transposes
        # here to do broadcasting from left to right (transposed), and then transpose back.
        # The result is an array the wind speed and wind direction dimensions on the left side
        # of the shape and the grid.template array on the right
        self.u_initial = (self.wind_speeds[None, :].T * wind_profile_plane.T).T * speed_ups
        self.v_initial = np.zeros(np.shape(self.u_initial), dtype=self.u_initial.dtype)
        self.w_initial = np.zeros(np.shape(self.u_initial), dtype=self.u_initial.dtype)

        self.u = self.u_initial.copy()
        self.v = self.v_initial.copy()
        self.w = self.w_initial.copy()

    def finalize(self, unsorted_indices):
        self.u = np.take_along_axis(self.u, unsorted_indices, axis=2)
        self.v = np.take_along_axis(self.v, unsorted_indices, axis=2)
        self.w = np.take_along_axis(self.w, unsorted_indices, axis=2)

    def calculate_speed_ups(self, het_map, x, y, z=None):
        if z is not None:
            # Calculate the 3-dimensional speed ups; reshape is needed as the generator adds an extra dimension
            speed_ups = np.reshape(
                [het_map[0][i](x[i:i+1,:,:,:,:], y[i:i+1,:,:,:,:], z[i:i+1,:,:,:,:]) for i in range(len(het_map[0]))],
                np.shape(x)
            )

            # If there are any points requested outside the user-defined area, use the
            # nearest-neighbor interplonat to determine those speed up values
            if np.isnan(speed_ups).any():
                idx_nan = np.where(np.isnan(speed_ups))
                speed_ups_out_of_region = np.reshape(
                    [het_map[1][i](x[i:i+1,:,:,:,:], y[i:i+1,:,:,:,:], z[i:i+1,:,:,:,:]) for i in range(len(het_map[1]))],
                    np.shape(x)
                )

                speed_ups[idx_nan] = speed_ups_out_of_region[idx_nan]

        else:
            # Calculate the 2-dimensional speed ups; reshape is needed as the generator adds an extra dimension
            speed_ups = np.reshape(
                [het_map[0][i](x[i:i+1,:,:,:,:], y[i:i+1,:,:,:,:]) for i in range(len(het_map[0]))],
                np.shape(x)
            )

            # If there are any points requested outside the user-defined area, use the
            # nearest-neighbor interplonat to determine those speed up values
            if np.isnan(speed_ups).any():
                idx_nan = np.where(np.isnan(speed_ups))
                speed_ups_out_of_region = np.reshape(
                    [het_map[1][i](x[i:i+1,:,:,:,:], y[i:i+1,:,:,:,:]) for i in range(len(het_map[1]))],
                    np.shape(x)
                )

                speed_ups[idx_nan] = speed_ups_out_of_region[idx_nan]

        return speed_ups
